from nshtrainer._checkpoint.loader import (
    BestCheckpointStrategyConfig as BestCheckpointStrategyConfig,
)
from nshtrainer._checkpoint.loader import (
    CheckpointLoadingConfig as CheckpointLoadingConfig,
)
from nshtrainer._checkpoint.loader import (
    CheckpointLoadingStrategyConfig as CheckpointLoadingStrategyConfig,
)
from nshtrainer._checkpoint.loader import (
    LastCheckpointStrategyConfig as LastCheckpointStrategyConfig,
)
from nshtrainer._checkpoint.loader import (
    UserProvidedPathCheckpointStrategyConfig as UserProvidedPathCheckpointStrategyConfig,
)
from nshtrainer._checkpoint.metadata import CheckpointMetadata as CheckpointMetadata
from nshtrainer._directory import DirectoryConfig as DirectoryConfig
from nshtrainer._hf_hub import (
    HuggingFaceHubAutoCreateConfig as HuggingFaceHubAutoCreateConfig,
)
from nshtrainer._hf_hub import HuggingFaceHubConfig as HuggingFaceHubConfig
from nshtrainer.callbacks.actsave import ActSaveConfig as ActSaveConfig
from nshtrainer.callbacks.base import CallbackConfigBase as CallbackConfigBase
from nshtrainer.callbacks.checkpoint._base import (
    BaseCheckpointCallbackConfig as BaseCheckpointCallbackConfig,
)
from nshtrainer.callbacks.checkpoint.best_checkpoint import (
    BestCheckpointCallbackConfig as BestCheckpointCallbackConfig,
)
from nshtrainer.callbacks.checkpoint.last_checkpoint import (
    LastCheckpointCallbackConfig as LastCheckpointCallbackConfig,
)
from nshtrainer.callbacks.checkpoint.on_exception_checkpoint import (
    OnExceptionCheckpointCallbackConfig as OnExceptionCheckpointCallbackConfig,
)
from nshtrainer.callbacks.debug_flag import (
    DebugFlagCallbackConfig as DebugFlagCallbackConfig,
)
from nshtrainer.callbacks.directory_setup import (
    DirectorySetupConfig as DirectorySetupConfig,
)
from nshtrainer.callbacks.early_stopping import (
    EarlyStoppingConfig as EarlyStoppingConfig,
)
from nshtrainer.callbacks.ema import EMAConfig as EMAConfig
from nshtrainer.callbacks.finite_checks import FiniteChecksConfig as FiniteChecksConfig
from nshtrainer.callbacks.gradient_skipping import (
    GradientSkippingConfig as GradientSkippingConfig,
)
from nshtrainer.callbacks.norm_logging import NormLoggingConfig as NormLoggingConfig
from nshtrainer.callbacks.print_table import (
    PrintTableMetricsConfig as PrintTableMetricsConfig,
)
from nshtrainer.callbacks.rlp_sanity_checks import (
    RLPSanityChecksConfig as RLPSanityChecksConfig,
)
from nshtrainer.callbacks.shared_parameters import (
    SharedParametersConfig as SharedParametersConfig,
)
from nshtrainer.callbacks.throughput_monitor import (
    ThroughputMonitorConfig as ThroughputMonitorConfig,
)
from nshtrainer.callbacks.timer import EpochTimerConfig as EpochTimerConfig
from nshtrainer.callbacks.wandb_upload_code import (
    WandbUploadCodeConfig as WandbUploadCodeConfig,
)
from nshtrainer.callbacks.wandb_watch import WandbWatchConfig as WandbWatchConfig
from nshtrainer.loggers._base import BaseLoggerConfig as BaseLoggerConfig
from nshtrainer.loggers.csv import CSVLoggerConfig as CSVLoggerConfig
from nshtrainer.loggers.tensorboard import (
    TensorboardLoggerConfig as TensorboardLoggerConfig,
)
from nshtrainer.loggers.wandb import WandbLoggerConfig as WandbLoggerConfig
from nshtrainer.lr_scheduler import LRSchedulerConfig as LRSchedulerConfig
from nshtrainer.lr_scheduler._base import LRSchedulerConfigBase as LRSchedulerConfigBase
from nshtrainer.lr_scheduler.linear_warmup_cosine import (
    DurationConfig as DurationConfig,
)
from nshtrainer.lr_scheduler.linear_warmup_cosine import (
    LinearWarmupCosineDecayLRSchedulerConfig as LinearWarmupCosineDecayLRSchedulerConfig,
)
from nshtrainer.lr_scheduler.reduce_lr_on_plateau import (
    ReduceLROnPlateauConfig as ReduceLROnPlateauConfig,
)
from nshtrainer.metrics._config import MetricConfig as MetricConfig
from nshtrainer.model.config import BaseConfig as BaseConfig
from nshtrainer.nn.mlp import MLPConfig as MLPConfig
from nshtrainer.nn.nonlinearity import BaseNonlinearityConfig as BaseNonlinearityConfig
from nshtrainer.nn.nonlinearity import ELUNonlinearityConfig as ELUNonlinearityConfig
from nshtrainer.nn.nonlinearity import GELUNonlinearityConfig as GELUNonlinearityConfig
from nshtrainer.nn.nonlinearity import (
    LeakyReLUNonlinearityConfig as LeakyReLUNonlinearityConfig,
)
from nshtrainer.nn.nonlinearity import MishNonlinearityConfig as MishNonlinearityConfig
from nshtrainer.nn.nonlinearity import NonlinearityConfig as NonlinearityConfig
from nshtrainer.nn.nonlinearity import PReLUConfig as PReLUConfig
from nshtrainer.nn.nonlinearity import ReLUNonlinearityConfig as ReLUNonlinearityConfig
from nshtrainer.nn.nonlinearity import (
    SigmoidNonlinearityConfig as SigmoidNonlinearityConfig,
)
from nshtrainer.nn.nonlinearity import SiLUNonlinearityConfig as SiLUNonlinearityConfig
from nshtrainer.nn.nonlinearity import (
    SoftmaxNonlinearityConfig as SoftmaxNonlinearityConfig,
)
from nshtrainer.nn.nonlinearity import (
    SoftplusNonlinearityConfig as SoftplusNonlinearityConfig,
)
from nshtrainer.nn.nonlinearity import (
    SoftsignNonlinearityConfig as SoftsignNonlinearityConfig,
)
from nshtrainer.nn.nonlinearity import (
    SwiGLUNonlinearityConfig as SwiGLUNonlinearityConfig,
)
from nshtrainer.nn.nonlinearity import (
    SwishNonlinearityConfig as SwishNonlinearityConfig,
)
from nshtrainer.nn.nonlinearity import TanhNonlinearityConfig as TanhNonlinearityConfig
from nshtrainer.optimizer import AdamWConfig as AdamWConfig
from nshtrainer.optimizer import OptimizerConfig as OptimizerConfig
from nshtrainer.optimizer import OptimizerConfigBase as OptimizerConfigBase
from nshtrainer.profiler._base import BaseProfilerConfig as BaseProfilerConfig
from nshtrainer.profiler.advanced import (
    AdvancedProfilerConfig as AdvancedProfilerConfig,
)
from nshtrainer.profiler.pytorch import PyTorchProfilerConfig as PyTorchProfilerConfig
from nshtrainer.profiler.simple import SimpleProfilerConfig as SimpleProfilerConfig
from nshtrainer.trainer._config import CallbackConfig as CallbackConfig
from nshtrainer.trainer._config import (
    CheckpointCallbackConfig as CheckpointCallbackConfig,
)
from nshtrainer.trainer._config import CheckpointSavingConfig as CheckpointSavingConfig
from nshtrainer.trainer._config import GradientClippingConfig as GradientClippingConfig
from nshtrainer.trainer._config import LoggerConfig as LoggerConfig
from nshtrainer.trainer._config import LoggingConfig as LoggingConfig
from nshtrainer.trainer._config import OptimizationConfig as OptimizationConfig
from nshtrainer.trainer._config import ProfilerConfig as ProfilerConfig
from nshtrainer.trainer._config import ReproducibilityConfig as ReproducibilityConfig
from nshtrainer.trainer._config import SanityCheckingConfig as SanityCheckingConfig
from nshtrainer.trainer._config import TrainerConfig as TrainerConfig
from nshtrainer.util._environment_info import (
    EnvironmentClassInformationConfig as EnvironmentClassInformationConfig,
)
from nshtrainer.util._environment_info import EnvironmentConfig as EnvironmentConfig
from nshtrainer.util._environment_info import (
    EnvironmentCUDAConfig as EnvironmentCUDAConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentGPUConfig as EnvironmentGPUConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentHardwareConfig as EnvironmentHardwareConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentLinuxEnvironmentConfig as EnvironmentLinuxEnvironmentConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentLSFInformationConfig as EnvironmentLSFInformationConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentPackageConfig as EnvironmentPackageConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentSLURMInformationConfig as EnvironmentSLURMInformationConfig,
)
from nshtrainer.util._environment_info import (
    EnvironmentSnapshotConfig as EnvironmentSnapshotConfig,
)
from nshtrainer.util._environment_info import GitRepositoryConfig as GitRepositoryConfig
from nshtrainer.util.config.dtype import DTypeConfig as DTypeConfig
from nshtrainer.util.config.duration import EpochsConfig as EpochsConfig
from nshtrainer.util.config.duration import StepsConfig as StepsConfig



import copy
import logging
import os
import string
import time
from collections.abc import Iterable
from pathlib import Path
from typing import Annotated, Any, ClassVar

import nshconfig as C
import numpy as np
import torch
from typing_extensions import Self

from .._directory import DirectoryConfig
from ..callbacks.base import CallbackConfigBase
from ..metrics import MetricConfig
from ..trainer._config import TrainerConfig
from ..util._environment_info import EnvironmentConfig

log = logging.getLogger(__name__)


class BaseConfig(C.Config):
    id: str = C.Field(default_factory=lambda: BaseConfig.generate_id())
    """ID of the run."""
    name: str | None = None
    """Run name."""
    name_parts: list[str] = []
    """A list of parts used to construct the run name. This is useful for constructing the run name dynamically."""
    project: str | None = None
    """Project name."""
    tags: list[str] = []
    """Tags for the run."""
    notes: list[str] = []
    """Human readable notes for the run."""

    debug: bool = False
    """Whether to run in debug mode. This will enable debug logging and enable debug code paths."""
    environment: Annotated[EnvironmentConfig, C.Field(repr=False)] = (
        EnvironmentConfig.empty()
    )
    """A snapshot of the current environment information (e.g. python version, slurm info, etc.). This is automatically populated by the run script."""

    directory: DirectoryConfig = DirectoryConfig()
    """Directory configuration options."""
    trainer: TrainerConfig = TrainerConfig()
    """PyTorch Lightning trainer configuration options. Check Lightning's `Trainer` documentation for more information."""

    primary_metric: MetricConfig | None = None
    """Primary metric configuration options. This is used in the following ways:
    - To determine the best model checkpoint to save with the ModelCheckpoint callback.
    - To monitor the primary metric during training and stop training based on the `early_stopping` configuration.
    - For the ReduceLROnPlateau scheduler.
    """

    meta: dict[str, Any] = {}
    """Additional metadata for this run. This can be used to store arbitrary data that is not part of the config schema."""

    @property
    def run_name(self) -> str:
        parts = self.name_parts.copy()
        if self.name is not None:
            parts = [self.name] + parts
        name = "-".join(parts)
        if not name:
            name = self.id
        return name

    def clone(self, with_new_id: bool = True) -> Self:
        c = copy.deepcopy(self)
        if with_new_id:
            c.id = BaseConfig.generate_id()
        return c

    def subdirectory(self, subdirectory: str) -> Path:
        return self.directory.resolve_subdirectory(self.id, subdirectory)

    # region Helper methods
    def fast_dev_run(self, value: int | bool = True, /):
        """
        Enables fast_dev_run mode for the trainer.
        This will run the training loop for a specified number of batches,
        if an integer is provided, or for a single batch if True is provided.
        """
        config = copy.deepcopy(self)
        config.trainer.fast_dev_run = value
        return config

    def with_project_root_(self, project_root: str | Path | os.PathLike) -> Self:
        """
        Set the project root directory for the trainer.

        Args:
            project_root (Path): The base directory to use.

        Returns:
            self: The current instance of the class.
        """
        self.directory.project_root = Path(project_root)
        return self

    def reset_(
        self,
        *,
        id: bool = True,
        basic: bool = True,
        project_root: bool = True,
        environment: bool = True,
        meta: bool = True,
    ):
        """
        Reset the configuration object to its initial state.

        Parameters:
        - id (bool): If True, generate a new ID for the configuration object.
        - basic (bool): If True, reset basic attributes like name, project, tags, and notes.
        - project_root (bool): If True, reset the directory configuration to its initial state.
        - environment (bool): If True, reset the environment configuration to its initial state.
        - meta (bool): If True, reset the meta dictionary to an empty dictionary.

        Returns:
        - self: The updated configuration object.

        """
        if id:
            self.id = self.generate_id()

        if basic:
            self.name = None
            self.name_parts = []
            self.project = None
            self.tags = []
            self.notes = []

        if project_root:
            self.directory = DirectoryConfig()

        if environment:
            self.environment = EnvironmentConfig.empty()

        if meta:
            self.meta = {}

        return self

    def concise_repr(self) -> str:
        """Get a concise representation of the configuration object."""

        def _truncate(s: str, max_len: int = 50):
            return s if len(s) <= max_len else f"{s[:max_len - 3]}..."

        cls_name = self.__class__.__name__

        parts: list[str] = []
        parts.append(f"name={self.run_name}")
        if self.project:
            parts.append(f"project={_truncate(self.project)}")

        return f"{cls_name}({', '.join(parts)})"

    # endregion

    # region Seeding

    _rng: ClassVar[np.random.Generator | None] = None

    @staticmethod
    def generate_id(*, length: int = 8) -> str:
        """
        Generate a random ID of specified length.

        """
        if (rng := BaseConfig._rng) is None:
            rng = np.random.default_rng()

        alphabet = list(string.ascii_lowercase + string.digits)

        id = "".join(rng.choice(alphabet) for _ in range(length))
        return id

    @staticmethod
    def set_seed(seed: int | None = None) -> None:
        """
        Set the seed for the random number generator.

        Args:
            seed (int | None, optional): The seed value to set. If None, a seed based on the current time will be used. Defaults to None.

        Returns:
            None
        """
        if seed is None:
            seed = int(time.time() * 1000)
        log.critical(f"Seeding BaseConfig with seed {seed}")
        BaseConfig._rng = np.random.default_rng(seed)

    # endregion

    @classmethod
    def from_checkpoint(
        cls,
        path: str | Path,
        *,
        hparams_key: str = "hyper_parameters",
    ):
        ckpt = torch.load(path)
        if (hparams := ckpt.get(hparams_key)) is None:
            raise ValueError(
                f"The checkpoint does not contain the `{hparams_key}` attribute. "
                "Are you sure this is a valid Lightning checkpoint?"
            )
        return cls.model_validate(hparams)

    def _nshtrainer_all_callback_configs(self) -> Iterable[CallbackConfigBase | None]:
        yield from self.trainer._nshtrainer_all_callback_configs()



import getpass
import importlib.metadata
import inspect
import logging
import os
import platform
import socket
import sys
from datetime import timedelta
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import nshconfig as C
import psutil
import torch
from packaging import version
from typing_extensions import Self

from .slurm import parse_slurm_node_list

if TYPE_CHECKING:
    from ..model.base import LightningModuleBase
    from ..model.config import BaseConfig


log = logging.getLogger(__name__)


class EnvironmentClassInformationConfig(C.Config):
    """Configuration for class information in the environment."""

    name: str | None = None
    """The name of the class."""

    module: str | None = None
    """The module where the class is defined."""

    full_name: str | None = None
    """The fully qualified name of the class."""

    file_path: Path | None = None
    """The file path where the class is defined."""

    source_file_path: Path | None = None
    """The source file path of the class, if available."""

    @classmethod
    def empty(cls):
        return cls(
            name=None,
            module=None,
            full_name=None,
            file_path=None,
            source_file_path=None,
        )

    @classmethod
    def from_class(cls, cls_: type):
        name = cls_.__name__
        module = cls_.__module__
        full_name = f"{cls_.__module__}.{cls_.__qualname__}"

        file_path = inspect.getfile(cls_)
        source_file_path = inspect.getsourcefile(cls_)
        return cls(
            name=name,
            module=module,
            full_name=full_name,
            file_path=Path(file_path),
            source_file_path=Path(source_file_path) if source_file_path else None,
        )

    @classmethod
    def from_instance(cls, instance: object):
        return cls.from_class(type(instance))


class EnvironmentSLURMInformationConfig(C.Config):
    """Configuration for SLURM environment information."""

    hostname: str | None = None
    """The hostname of the current node."""

    hostnames: list[str] | None = None
    """List of hostnames for all nodes in the job."""

    job_id: str | None = None
    """The SLURM job ID."""

    raw_job_id: str | None = None
    """The raw SLURM job ID."""

    array_job_id: str | None = None
    """The SLURM array job ID, if applicable."""

    array_task_id: str | None = None
    """The SLURM array task ID, if applicable."""

    num_tasks: int | None = None
    """The number of tasks in the SLURM job."""

    num_nodes: int | None = None
    """The number of nodes in the SLURM job."""

    node: str | int | None = None
    """The node ID or name."""

    global_rank: int | None = None
    """The global rank of the current process."""

    local_rank: int | None = None
    """The local rank of the current process within its node."""

    @classmethod
    def empty(cls):
        return cls(
            hostname=None,
            hostnames=None,
            job_id=None,
            raw_job_id=None,
            array_job_id=None,
            array_task_id=None,
            num_tasks=None,
            num_nodes=None,
            node=None,
            global_rank=None,
            local_rank=None,
        )

    @classmethod
    def from_current_environment(cls):
        try:
            from lightning.fabric.plugins.environments.slurm import SLURMEnvironment

            if not SLURMEnvironment.detect():
                return None

            hostname = socket.gethostname()
            hostnames = [hostname]
            if node_list := os.environ.get("SLURM_JOB_NODELIST", ""):
                hostnames = parse_slurm_node_list(node_list)

            raw_job_id = os.environ["SLURM_JOB_ID"]
            job_id = raw_job_id
            array_job_id = os.environ.get("SLURM_ARRAY_JOB_ID")
            array_task_id = os.environ.get("SLURM_ARRAY_TASK_ID")
            if array_job_id and array_task_id:
                job_id = f"{array_job_id}_{array_task_id}"

            num_tasks = int(os.environ["SLURM_NTASKS"])
            num_nodes = int(os.environ["SLURM_JOB_NUM_NODES"])

            node_id = os.environ.get("SLURM_NODEID")

            global_rank = int(os.environ["SLURM_PROCID"])
            local_rank = int(os.environ["SLURM_LOCALID"])

            return cls(
                hostname=hostname,
                hostnames=hostnames,
                job_id=job_id,
                raw_job_id=raw_job_id,
                array_job_id=array_job_id,
                array_task_id=array_task_id,
                num_tasks=num_tasks,
                num_nodes=num_nodes,
                node=node_id,
                global_rank=global_rank,
                local_rank=local_rank,
            )
        except (ImportError, RuntimeError, ValueError, KeyError):
            return None


class EnvironmentLSFInformationConfig(C.Config):
    """Configuration for LSF environment information."""

    hostname: str | None = None
    """The hostname of the current node."""

    hostnames: list[str] | None = None
    """List of hostnames for all nodes in the job."""

    job_id: str | None = None
    """The LSF job ID."""

    array_job_id: str | None = None
    """The LSF array job ID, if applicable."""

    array_task_id: str | None = None
    """The LSF array task ID, if applicable."""

    num_tasks: int | None = None
    """The number of tasks in the LSF job."""

    num_nodes: int | None = None
    """The number of nodes in the LSF job."""

    node: str | int | None = None
    """The node ID or name."""

    global_rank: int | None = None
    """The global rank of the current process."""

    local_rank: int | None = None
    """The local rank of the current process within its node."""

    @classmethod
    def empty(cls):
        return cls(
            hostname=None,
            hostnames=None,
            job_id=None,
            array_job_id=None,
            array_task_id=None,
            num_tasks=None,
            num_nodes=None,
            node=None,
            global_rank=None,
            local_rank=None,
        )

    @classmethod
    def from_current_environment(cls):
        try:
            import os
            import socket

            hostname = socket.gethostname()
            hostnames = [hostname]
            if node_list := os.environ.get("LSB_HOSTS", ""):
                hostnames = node_list.split()

            job_id = os.environ["LSB_JOBID"]
            array_job_id = os.environ.get("LSB_JOBINDEX")
            array_task_id = os.environ.get("LSB_JOBINDEX")

            num_tasks = int(os.environ.get("LSB_DJOB_NUMPROC", 1))
            num_nodes = len(set(hostnames))

            node_id = (
                os.environ.get("LSB_HOSTS", "").split().index(hostname)
                if "LSB_HOSTS" in os.environ
                else None
            )

            # LSF doesn't have direct equivalents for global_rank and local_rank
            # You might need to calculate these based on your specific setup
            global_rank = int(os.environ.get("PMI_RANK", 0))
            local_rank = int(os.environ.get("LSB_RANK", 0))

            return cls(
                hostname=hostname,
                hostnames=hostnames,
                job_id=job_id,
                array_job_id=array_job_id,
                array_task_id=array_task_id,
                num_tasks=num_tasks,
                num_nodes=num_nodes,
                node=node_id,
                global_rank=global_rank,
                local_rank=local_rank,
            )
        except (ImportError, RuntimeError, ValueError, KeyError):
            return None


class EnvironmentLinuxEnvironmentConfig(C.Config):
    """Configuration for Linux environment information."""

    user: str | None = None
    """The current user."""

    hostname: str | None = None
    """The hostname of the machine."""

    system: str | None = None
    """The operating system name."""

    release: str | None = None
    """The operating system release."""

    version: str | None = None
    """The operating system version."""

    machine: str | None = None
    """The machine type."""

    processor: str | None = None
    """The processor type."""

    cpu_count: int | None = None
    """The number of CPUs."""

    memory: int | None = None
    """The total system memory in bytes."""

    uptime: timedelta | None = None
    """The system uptime."""

    boot_time: float | None = None
    """The system boot time as a timestamp."""

    load_avg: tuple[float, float, float] | None = None
    """The system load average (1, 5, and 15 minutes)."""

    @classmethod
    def empty(cls):
        return cls(
            user=None,
            hostname=None,
            system=None,
            release=None,
            version=None,
            machine=None,
            processor=None,
            cpu_count=None,
            memory=None,
            uptime=None,
            boot_time=None,
            load_avg=None,
        )

    @classmethod
    def from_current_environment(cls):
        return cls(
            user=getpass.getuser(),
            hostname=platform.node(),
            system=platform.system(),
            release=platform.release(),
            version=platform.version(),
            machine=platform.machine(),
            processor=platform.processor(),
            cpu_count=os.cpu_count(),
            memory=psutil.virtual_memory().total,
            uptime=timedelta(seconds=psutil.boot_time()),
            boot_time=psutil.boot_time(),
            load_avg=os.getloadavg(),
        )


class EnvironmentSnapshotConfig(C.Config):
    """Configuration for environment snapshot information."""

    snapshot_dir: Path | None = None
    """The directory where the snapshot is stored."""

    modules: list[str] | None = None
    """List of modules included in the snapshot."""

    @classmethod
    def empty(cls):
        return cls(snapshot_dir=None, modules=None)

    @classmethod
    def from_current_environment(cls):
        draft = cls.draft()
        if snapshot_dir := os.environ.get("NSHRUNNER_SNAPSHOT_DIR"):
            draft.snapshot_dir = Path(snapshot_dir)
        if modules := os.environ.get("NSHRUNNER_SNAPSHOT_MODULES"):
            draft.modules = modules.split(",")
        return draft.finalize()


class EnvironmentPackageConfig(C.Config):
    """Configuration for Python package information."""

    name: str | None = None
    """The name of the package."""

    version: str | None = None
    """The version of the package."""

    path: Path | None = None
    """The installation path of the package."""

    summary: str | None = None
    """A brief summary of the package."""

    author: str | None = None
    """The author of the package."""

    license: str | None = None
    """The license of the package."""

    requires: list[str] | None = None
    """List of package dependencies."""

    @classmethod
    def empty(cls):
        return cls(
            name=None,
            version=None,
            path=None,
            summary=None,
            author=None,
            license=None,
            requires=None,
        )

    @classmethod
    def from_current_environment(cls):
        python_packages: dict[str, Self] = {}
        try:
            for dist in importlib.metadata.distributions():
                try:
                    # Get package metadata
                    metadata = dist.metadata

                    # Parse the version, stripping any local version identifier
                    pkg_version = version.parse(dist.version)
                    clean_version = (
                        f"{pkg_version.major}.{pkg_version.minor}.{pkg_version.micro}"
                    )

                    # Get requirements
                    requires = []
                    for req in dist.requires or []:
                        try:
                            requires.append(str(req))
                        except ValueError:
                            # If there's an invalid requirement, we'll skip it
                            log.warning(
                                f"Skipping invalid requirement for {dist.name}: {req}"
                            )

                    python_packages[dist.name] = cls(
                        name=dist.name,
                        version=clean_version,
                        path=Path(str(f)) if (f := dist.locate_file("")) else None,
                        summary=metadata["Summary"] if "Summary" in metadata else None,
                        author=metadata["Author"] if "Author" in metadata else None,
                        license=metadata["License"] if "License" in metadata else None,
                        requires=requires,
                    )
                except Exception:
                    log.warning(f"Error processing package {dist.name}", exc_info=True)

        except ImportError:
            log.warning(
                "importlib.metadata not available, skipping package information"
            )

        return python_packages


class EnvironmentGPUConfig(C.Config):
    """Configuration for individual GPU information."""

    name: str | None = None
    """Name of the GPU."""

    total_memory: int | None = None
    """Total memory of the GPU in bytes."""

    major: int | None = None
    """Major version of CUDA capability."""

    minor: int | None = None
    """Minor version of CUDA capability."""

    multi_processor_count: int | None = None
    """Number of multiprocessors on the GPU."""

    @classmethod
    def empty(cls):
        return cls(
            name=None,
            total_memory=None,
            major=None,
            minor=None,
            multi_processor_count=None,
        )


class EnvironmentCUDAConfig(C.Config):
    """Configuration for CUDA environment information."""

    is_available: bool | None = None
    """Whether CUDA is available."""

    version: str | None = None
    """CUDA version."""

    cudnn_version: int | None = None
    """cuDNN version."""

    @classmethod
    def empty(cls):
        return cls(is_available=None, version=None, cudnn_version=None)


class EnvironmentHardwareConfig(C.Config):
    """Configuration for hardware information."""

    cpu_count_physical: int | None = None
    """Number of physical CPU cores."""

    cpu_count_logical: int | None = None
    """Number of logical CPU cores."""

    cpu_frequency_current: float | None = None
    """Current CPU frequency in MHz."""

    cpu_frequency_min: float | None = None
    """Minimum CPU frequency in MHz."""

    cpu_frequency_max: float | None = None
    """Maximum CPU frequency in MHz."""

    ram_total: int | None = None
    """Total RAM in bytes."""

    ram_available: int | None = None
    """Available RAM in bytes."""

    disk_total: int | None = None
    """Total disk space in bytes."""

    disk_used: int | None = None
    """Used disk space in bytes."""

    disk_free: int | None = None
    """Free disk space in bytes."""

    gpu_count: int | None = None
    """Number of GPUs available."""

    gpus: list[EnvironmentGPUConfig] | None = None
    """List of GPU configurations."""

    cuda: EnvironmentCUDAConfig | None = None
    """CUDA environment configuration."""

    @classmethod
    def empty(cls):
        return cls(
            cpu_count_physical=None,
            cpu_count_logical=None,
            cpu_frequency_current=None,
            cpu_frequency_min=None,
            cpu_frequency_max=None,
            ram_total=None,
            ram_available=None,
            disk_total=None,
            disk_used=None,
            disk_free=None,
            gpu_count=None,
            gpus=None,
            cuda=None,
        )

    @classmethod
    def from_current_environment(cls):
        draft = cls.draft()

        # CPU information
        draft.cpu_count_physical = psutil.cpu_count(logical=False)
        draft.cpu_count_logical = psutil.cpu_count(logical=True)
        cpu_freq = psutil.cpu_freq()
        if cpu_freq:
            draft.cpu_frequency_current = cpu_freq.current
            draft.cpu_frequency_min = cpu_freq.min
            draft.cpu_frequency_max = cpu_freq.max

        # RAM information
        ram = psutil.virtual_memory()
        draft.ram_total = ram.total
        draft.ram_available = ram.available

        # Disk information
        disk = psutil.disk_usage("/")
        draft.disk_total = disk.total
        draft.disk_used = disk.used
        draft.disk_free = disk.free

        # GPU and CUDA information
        draft.cuda = EnvironmentCUDAConfig(
            is_available=torch.cuda.is_available(),
            version=cast(Any, torch).version.cuda,
            cudnn_version=torch.backends.cudnn.version()
            if torch.backends.cudnn.is_available()
            else None,
        )

        if draft.cuda.is_available:
            draft.gpu_count = torch.cuda.device_count()
            draft.gpus = []
            for i in range(draft.gpu_count):
                gpu_props = torch.cuda.get_device_properties(i)
                gpu_config = EnvironmentGPUConfig(
                    name=gpu_props.name,
                    total_memory=gpu_props.total_memory,
                    major=gpu_props.major,
                    minor=gpu_props.minor,
                    multi_processor_count=gpu_props.multi_processor_count,
                )
                draft.gpus.append(gpu_config)

        return draft.finalize()


class GitRepositoryConfig(C.Config):
    """Configuration for Git repository information."""

    is_git_repo: bool | None = None
    """Whether the current directory is a Git repository."""

    branch: str | None = None
    """The current Git branch."""

    commit_hash: str | None = None
    """The current commit hash."""

    commit_message: str | None = None
    """The current commit message."""

    author: str | None = None
    """The author of the current commit."""

    commit_date: str | None = None
    """The date of the current commit."""

    remote_url: str | None = None
    """The URL of the remote repository."""

    is_dirty: bool | None = None
    """Whether there are uncommitted changes."""

    @classmethod
    def empty(cls):
        return cls(
            is_git_repo=None,
            branch=None,
            commit_hash=None,
            commit_message=None,
            author=None,
            commit_date=None,
            remote_url=None,
            is_dirty=None,
        )

    @classmethod
    def from_current_directory(cls):
        try:
            import git
        except ImportError:
            return cls()

        draft = cls.draft()
        try:
            repo = git.Repo(os.getcwd(), search_parent_directories=True)
            draft.is_git_repo = True
            draft.branch = repo.active_branch.name
            commit = repo.head.commit
            draft.commit_hash = commit.hexsha

            # Handle both str and bytes for commit message
            if isinstance(commit.message, str):
                draft.commit_message = commit.message.strip()
            elif isinstance(commit.message, bytes):
                draft.commit_message = commit.message.decode(
                    "utf-8", errors="replace"
                ).strip()
            else:
                draft.commit_message = str(commit.message).strip()

            draft.author = f"{commit.author.name} <{commit.author.email}>"
            draft.commit_date = commit.committed_datetime.isoformat()
            if repo.remotes:
                draft.remote_url = repo.remotes.origin.url
            draft.is_dirty = repo.is_dirty()
        except git.InvalidGitRepositoryError:
            draft.is_git_repo = False
        except Exception:
            log.warning("Failed to get Git repository information", exc_info=True)
            draft.is_git_repo = None

        return draft.finalize()


class EnvironmentConfig(C.Config):
    """Configuration for the overall environment."""

    cwd: Path | None = None
    """The current working directory."""

    snapshot: EnvironmentSnapshotConfig | None = None
    """The environment snapshot configuration."""

    python_executable: Path | None = None
    """The path to the Python executable."""

    python_path: list[Path] | None = None
    """The Python path."""

    python_version: str | None = None
    """The Python version."""

    python_packages: dict[str, EnvironmentPackageConfig] | None = None
    """A mapping of package names to their configurations."""

    config: EnvironmentClassInformationConfig | None = None
    """The configuration class information."""

    model: EnvironmentClassInformationConfig | None = None
    """The Lightning module class information."""

    linux: EnvironmentLinuxEnvironmentConfig | None = None
    """The Linux environment information."""

    hardware: EnvironmentHardwareConfig | None = None
    """Hardware configuration information."""

    slurm: EnvironmentSLURMInformationConfig | None = None
    """The SLURM environment information."""

    lsf: EnvironmentLSFInformationConfig | None = None
    """The LSF environment information."""

    base_dir: Path | None = None
    """The base directory for the run."""

    log_dir: Path | None = None
    """The directory for logs."""

    checkpoint_dir: Path | None = None
    """The directory for checkpoints."""

    stdio_dir: Path | None = None
    """The directory for standard input/output files."""

    seed: int | None = None
    """The global random seed."""

    seed_workers: bool | None = None
    """Whether to seed workers."""

    git: GitRepositoryConfig | None = None
    """Git repository information."""

    @classmethod
    def empty(cls):
        return cls(
            cwd=None,
            snapshot=None,
            python_executable=None,
            python_path=None,
            python_version=None,
            python_packages=None,
            config=None,
            model=None,
            linux=None,
            hardware=None,
            slurm=None,
            lsf=None,
            base_dir=None,
            log_dir=None,
            checkpoint_dir=None,
            stdio_dir=None,
            seed=None,
            seed_workers=None,
            git=None,
        )

    @classmethod
    def from_current_environment(
        cls,
        root_config: "BaseConfig",
        model: "LightningModuleBase",
    ):
        draft = cls.draft()
        draft.cwd = Path(os.getcwd())
        draft.python_executable = Path(sys.executable)
        draft.python_path = [Path(path) for path in sys.path]
        draft.python_version = sys.version
        draft.python_packages = EnvironmentPackageConfig.from_current_environment()
        draft.config = EnvironmentClassInformationConfig.from_instance(root_config)
        draft.model = EnvironmentClassInformationConfig.from_instance(model)
        draft.linux = EnvironmentLinuxEnvironmentConfig.from_current_environment()
        draft.hardware = EnvironmentHardwareConfig.from_current_environment()
        draft.slurm = EnvironmentSLURMInformationConfig.from_current_environment()
        draft.lsf = EnvironmentLSFInformationConfig.from_current_environment()
        draft.base_dir = root_config.directory.resolve_run_root_directory(
            root_config.id
        )
        draft.log_dir = root_config.directory.resolve_subdirectory(
            root_config.id, "log"
        )
        draft.checkpoint_dir = root_config.directory.resolve_subdirectory(
            root_config.id, "checkpoint"
        )
        draft.stdio_dir = root_config.directory.resolve_subdirectory(
            root_config.id, "stdio"
        )
        draft.seed = (
            int(seed_str) if (seed_str := os.environ.get("PL_GLOBAL_SEED")) else None
        )
        draft.seed_workers = (
            bool(int(seed_everything))
            if (seed_everything := os.environ.get("PL_SEED_WORKERS"))
            else None
        )
        draft.snapshot = EnvironmentSnapshotConfig.from_current_environment()
        draft.git = GitRepositoryConfig.from_current_directory()
        return draft.finalize()



import contextlib
import copy
import threading
from collections.abc import Callable, Iterable
from typing import Any, Literal, overload

import lightning.pytorch as pl
import torch
from lightning.pytorch import Callback
from lightning.pytorch.utilities.exceptions import MisconfigurationException
from typing_extensions import override

from .base import CallbackConfigBase


class EMA(Callback):
    """
    Implements Exponential Moving Averaging (EMA).

    When training a model, this callback will maintain moving averages of the trained parameters.
    When evaluating, we use the moving averages copy of the trained parameters.
    When saving, we save an additional set of parameters with the prefix `ema`.

    Args:
        decay: The exponential decay used when calculating the moving average. Has to be between 0-1.
        validate_original_weights: Validate the original weights, as apposed to the EMA weights.
        every_n_steps: Apply EMA every N steps.
        cpu_offload: Offload weights to CPU.
    """

    @override
    def __init__(
        self,
        decay: float,
        validate_original_weights: bool = False,
        every_n_steps: int = 1,
        cpu_offload: bool = False,
    ):
        if not (0 <= decay <= 1):
            raise MisconfigurationException("EMA decay value must be between 0 and 1")
        self.decay = decay
        self.validate_original_weights = validate_original_weights
        self.every_n_steps = every_n_steps
        self.cpu_offload = cpu_offload

    @override
    def on_fit_start(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        device = pl_module.device if not self.cpu_offload else torch.device("cpu")
        trainer.optimizers = [
            EMAOptimizer(
                optim,
                device=device,
                decay=self.decay,
                every_n_steps=self.every_n_steps,
                current_step=trainer.global_step,
            )
            for optim in trainer.optimizers
            if not isinstance(optim, EMAOptimizer)
        ]

    @override
    def on_validation_start(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        if self._should_validate_ema_weights(trainer):
            self.swap_model_weights(trainer)

    @override
    def on_validation_end(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        if self._should_validate_ema_weights(trainer):
            self.swap_model_weights(trainer)

    @override
    def on_test_start(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        if self._should_validate_ema_weights(trainer):
            self.swap_model_weights(trainer)

    @override
    def on_test_end(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        if self._should_validate_ema_weights(trainer):
            self.swap_model_weights(trainer)

    def _should_validate_ema_weights(self, trainer: "pl.Trainer") -> bool:
        return not self.validate_original_weights and self._ema_initialized(trainer)

    def _ema_initialized(self, trainer: "pl.Trainer") -> bool:
        return any(
            isinstance(optimizer, EMAOptimizer) for optimizer in trainer.optimizers
        )

    def swap_model_weights(self, trainer: "pl.Trainer", saving_ema_model: bool = False):
        for optimizer in trainer.optimizers:
            assert isinstance(optimizer, EMAOptimizer)
            optimizer.switch_main_parameter_weights(saving_ema_model)

    @contextlib.contextmanager
    def save_ema_model(self, trainer: "pl.Trainer"):
        """
        Saves an EMA copy of the model + EMA optimizer states for resume.
        """
        self.swap_model_weights(trainer, saving_ema_model=True)
        try:
            yield
        finally:
            self.swap_model_weights(trainer, saving_ema_model=False)

    @contextlib.contextmanager
    def save_original_optimizer_state(self, trainer: "pl.Trainer"):
        for optimizer in trainer.optimizers:
            assert isinstance(optimizer, EMAOptimizer)
            optimizer.save_original_optimizer_state = True
        try:
            yield
        finally:
            for optimizer in trainer.optimizers:
                assert isinstance(optimizer, EMAOptimizer)
                optimizer.save_original_optimizer_state = False


@torch.no_grad()
def ema_update(ema_model_tuple, current_model_tuple, decay):
    torch._foreach_mul_(ema_model_tuple, decay)
    torch._foreach_add_(
        ema_model_tuple,
        current_model_tuple,
        alpha=(1.0 - decay),
    )


def run_ema_update_cpu(
    ema_model_tuple, current_model_tuple, decay, pre_sync_stream=None
):
    if pre_sync_stream is not None:
        pre_sync_stream.synchronize()

    ema_update(ema_model_tuple, current_model_tuple, decay)


class EMAOptimizer(torch.optim.Optimizer):
    r"""
    EMAOptimizer is a wrapper for torch.optim.Optimizer that computes
    Exponential Moving Average of parameters registered in the optimizer.

    EMA parameters are automatically updated after every step of the optimizer
    with the following formula:

        ema_weight = decay * ema_weight + (1 - decay) * training_weight

    To access EMA parameters, use ``swap_ema_weights()`` context manager to
    perform a temporary in-place swap of regular parameters with EMA
    parameters.

    Notes:
        - EMAOptimizer is not compatible with APEX AMP O2.

    Args:
        optimizer (torch.optim.Optimizer): optimizer to wrap
        device (torch.device): device for EMA parameters
        decay (float): decay factor

    Returns:
        returns an instance of torch.optim.Optimizer that computes EMA of
        parameters

    Example:
        model = Model().to(device)
        opt = torch.optim.Adam(model.parameters())

        opt = EMAOptimizer(opt, device, 0.9999)

        for epoch in range(epochs):
            training_loop(model, opt)

            regular_eval_accuracy = evaluate(model)

            with opt.swap_ema_weights():
                ema_eval_accuracy = evaluate(model)
    """

    stream: Any | None

    @override
    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        device: torch.device,
        decay: float = 0.9999,
        every_n_steps: int = 1,
        current_step: int = 0,
    ):
        self.optimizer = optimizer
        self.decay = decay
        self.device = device
        self.current_step = current_step
        self.every_n_steps = every_n_steps
        self.save_original_optimizer_state = False

        self.first_iteration = True
        self.rebuild_ema_params = True
        self.stream = None
        self.thread = None

        self.ema_params = ()
        self.in_saving_ema_model_context = False

    def all_parameters(self) -> Iterable[torch.Tensor]:
        return (param for group in self.param_groups for param in group["params"])

    @overload
    def step(self, closure: None = ...) -> None: ...

    @overload
    def step(self, closure: Callable[[], float]) -> float: ...

    @override
    def step(self, closure: Callable[[], float] | None = None) -> float | None:
        self.join()

        if self.first_iteration:
            if any(p.is_cuda for p in self.all_parameters()):
                self.stream = torch.cuda.Stream()

            self.first_iteration = False

        if self.rebuild_ema_params:
            opt_params = list(self.all_parameters())

            self.ema_params += tuple(
                copy.deepcopy(param.data.detach()).to(self.device)
                for param in opt_params[len(self.ema_params) :]
            )
            self.rebuild_ema_params = False

        loss = self.optimizer.step(closure)

        if self._should_update_at_step():
            self.update()
        self.current_step += 1
        return loss

    def _should_update_at_step(self) -> bool:
        return self.current_step % self.every_n_steps == 0

    @torch.no_grad()
    def update(self):
        if self.stream is not None:
            self.stream.wait_stream(torch.cuda.current_stream())

        with torch.cuda.stream(self.stream):
            current_model_state = tuple(
                param.data.to(self.device, non_blocking=True)
                for param in self.all_parameters()
            )

            if self.device.type == "cuda":
                ema_update(self.ema_params, current_model_state, self.decay)

        if self.device.type == "cpu":
            self.thread = threading.Thread(
                target=run_ema_update_cpu,
                args=(
                    self.ema_params,
                    current_model_state,
                    self.decay,
                    self.stream,
                ),
            )
            self.thread.start()

    def swap_tensors(self, tensor1, tensor2):
        tmp = torch.empty_like(tensor1)
        tmp.copy_(tensor1)
        tensor1.copy_(tensor2)
        tensor2.copy_(tmp)

    def switch_main_parameter_weights(self, saving_ema_model: bool = False):
        self.join()
        self.in_saving_ema_model_context = saving_ema_model
        for param, ema_param in zip(self.all_parameters(), self.ema_params):
            self.swap_tensors(param.data, ema_param)

    @contextlib.contextmanager
    def swap_ema_weights(self, enabled: bool = True):
        r"""
        A context manager to in-place swap regular parameters with EMA
        parameters.
        It swaps back to the original regular parameters on context manager
        exit.

        Args:
            enabled (bool): whether the swap should be performed
        """

        if enabled:
            self.switch_main_parameter_weights()
        try:
            yield
        finally:
            if enabled:
                self.switch_main_parameter_weights()

    def __getattr__(self, name):
        return getattr(self.optimizer, name)

    def join(self):
        if self.stream is not None:
            self.stream.synchronize()

        if self.thread is not None:
            self.thread.join()

    @override
    def state_dict(self):
        self.join()

        if self.save_original_optimizer_state:
            return self.optimizer.state_dict()

        # if we are in the context of saving an EMA model, the EMA weights are in the modules' actual weights
        ema_params = (
            self.ema_params
            if not self.in_saving_ema_model_context
            else list(self.all_parameters())
        )
        state_dict = {
            "opt": self.optimizer.state_dict(),
            "ema": ema_params,
            "current_step": self.current_step,
            "decay": self.decay,
            "every_n_steps": self.every_n_steps,
        }
        return state_dict

    @override
    def load_state_dict(self, state_dict):
        self.join()

        self.optimizer.load_state_dict(state_dict["opt"])
        self.ema_params = tuple(
            param.to(self.device) for param in copy.deepcopy(state_dict["ema"])
        )
        self.current_step = state_dict["current_step"]
        self.decay = state_dict["decay"]
        self.every_n_steps = state_dict["every_n_steps"]
        self.rebuild_ema_params = False

    @override
    def add_param_group(self, param_group):
        self.optimizer.add_param_group(param_group)
        self.rebuild_ema_params = True


class EMAConfig(CallbackConfigBase):
    name: Literal["ema"] = "ema"

    decay: float
    """The exponential decay used when calculating the moving average. Has to be between 0-1."""

    validate_original_weights: bool = False
    """Validate the original weights, as apposed to the EMA weights."""

    every_n_steps: int = 1
    """Apply EMA every N steps."""

    cpu_offload: bool = False
    """Offload weights to CPU."""

    @override
    def create_callbacks(self, root_config):
        yield EMA(
            decay=self.decay,
            validate_original_weights=self.validate_original_weights,
            every_n_steps=self.every_n_steps,
            cpu_offload=self.cpu_offload,
        )



import logging

import lightning.fabric.utilities.seed as LS

log = logging.getLogger(__name__)


def seed_everything(seed: int | None, *, workers: bool = False):
    seed = LS.seed_everything(seed, workers=workers)
    log.critical(f"Set global seed to {seed}.")
    return seed



import hashlib
import logging
import os
import platform
import shutil
from pathlib import Path
from typing import TypeAlias

log = logging.getLogger(__name__)

_Path: TypeAlias = str | Path | os.PathLike


def get_relative_path(source: _Path, destination: _Path):
    # Get the absolute paths
    source = os.path.abspath(source)
    destination = os.path.abspath(destination)

    # Split the paths into components
    source_parts = source.split(os.sep)
    destination_parts = destination.split(os.sep)

    # Find the point where the paths diverge
    i = 0
    for i in range(min(len(source_parts), len(destination_parts))):
        if source_parts[i] != destination_parts[i]:
            break
    else:
        i += 1

    # Build the relative path
    up = os.sep.join([".." for _ in range(len(source_parts) - i - 1)])
    down = os.sep.join(destination_parts[i:])

    return Path(os.path.normpath(os.path.join(up, down)))


def find_symlinks(
    target_file: _Path,
    *search_directories: _Path,
    glob_pattern: str = "*",
):
    target_file = Path(target_file).resolve()
    symlinks: list[Path] = []

    for search_directory in search_directories:
        search_directory = Path(search_directory)
        for path in search_directory.rglob(glob_pattern):
            if path.is_symlink():
                try:
                    link_target = path.resolve()
                    if link_target.samefile(target_file):
                        symlinks.append(path)
                except FileNotFoundError:
                    # Handle broken symlinks
                    pass

    return symlinks


def compute_file_checksum(file_path: Path) -> str:
    """
    Calculate the SHA256 checksum of a file.

    Args:
        file_path (Path): The path to the file.

    Returns:
        str: The hexadecimal representation of the file's SHA256 checksum.
    """
    sha256_hash = hashlib.sha256()
    with file_path.open("rb") as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()


def try_symlink_or_copy(
    file_path: Path,
    link_path: Path,
    target_is_directory: bool = False,
    relative: bool = True,
    remove_existing: bool = True,
):
    """
    Symlinks on Unix, copies on Windows.
    """

    # If the link already exists, remove it
    if remove_existing:
        try:
            if link_path.exists():
                if link_path.is_dir():
                    shutil.rmtree(link_path)
                else:
                    link_path.unlink(missing_ok=True)
        except Exception:
            log.warning(f"Failed to remove {link_path}", exc_info=True)
        else:
            log.debug(f"Removed {link_path=}")

    symlink_target = get_relative_path(link_path, file_path) if relative else file_path
    try:
        if platform.system() == "Windows":
            if target_is_directory:
                shutil.copytree(file_path, link_path)
            else:
                shutil.copy(file_path, link_path)
        else:
            link_path.symlink_to(
                symlink_target, target_is_directory=target_is_directory
            )
    except Exception:
        log.warning(
            f"Failed to create symlink or copy {file_path} to {link_path}",
            exc_info=True,
        )
        return False
    else:
        log.debug(f"Created symlink or copied {file_path} to {link_path}")
        return True



import math
from typing import Annotated, Literal

import nshconfig as C


class StepsConfig(C.Config):
    kind: Literal["steps"] = "steps"

    value: Annotated[int, C.Field(ge=0)]
    """Number of steps."""

    def to_steps(self, steps_per_epoch: int):
        return self


class EpochsConfig(C.Config):
    kind: Literal["epochs"] = "epochs"

    value: Annotated[int | float, C.Field(ge=0)]
    """Number of epochs."""

    def to_steps(self, steps_per_epoch: int):
        value = self.value * steps_per_epoch
        if not isinstance(value, int):
            value = int(math.ceil(value))

        return StepsConfig(value=value)


DurationConfig = Annotated[StepsConfig | EpochsConfig, C.Field(discriminator="kind")]



import logging
import time
from typing import Any, Literal

from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.callbacks import Callback
from lightning.pytorch.utilities.types import STEP_OUTPUT
from typing_extensions import override

from .base import CallbackConfigBase

log = logging.getLogger(__name__)


class EpochTimer(Callback):
    def __init__(self):
        super().__init__()

        self._start_time: dict[str, float] = {}
        self._elapsed_time: dict[str, float] = {}
        self._total_batches: dict[str, int] = {}

    @override
    def on_train_epoch_start(
        self, trainer: "Trainer", pl_module: "LightningModule"
    ) -> None:
        self._start_time["train"] = time.monotonic()
        self._total_batches["train"] = 0

    @override
    def on_train_batch_end(
        self,
        trainer: "Trainer",
        pl_module: "LightningModule",
        outputs: STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
    ) -> None:
        self._total_batches["train"] += 1

    @override
    def on_train_epoch_end(
        self, trainer: "Trainer", pl_module: "LightningModule"
    ) -> None:
        self._elapsed_time["train"] = time.monotonic() - self._start_time["train"]
        if trainer.is_global_zero:
            self._log_epoch_info("train")

    @override
    def on_validation_epoch_start(
        self, trainer: "Trainer", pl_module: "LightningModule"
    ) -> None:
        self._start_time["val"] = time.monotonic()
        self._total_batches["val"] = 0

    @override
    def on_validation_batch_end(
        self,
        trainer: "Trainer",
        pl_module: "LightningModule",
        outputs: STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        self._total_batches["val"] += 1

    @override
    def on_validation_epoch_end(
        self, trainer: "Trainer", pl_module: "LightningModule"
    ) -> None:
        self._elapsed_time["val"] = time.monotonic() - self._start_time["val"]
        if trainer.is_global_zero:
            self._log_epoch_info("val")

    @override
    def on_test_epoch_start(
        self, trainer: "Trainer", pl_module: "LightningModule"
    ) -> None:
        self._start_time["test"] = time.monotonic()
        self._total_batches["test"] = 0

    @override
    def on_test_batch_end(
        self,
        trainer: "Trainer",
        pl_module: "LightningModule",
        outputs: STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        self._total_batches["test"] += 1

    @override
    def on_test_epoch_end(
        self, trainer: "Trainer", pl_module: "LightningModule"
    ) -> None:
        self._elapsed_time["test"] = time.monotonic() - self._start_time["test"]
        if trainer.is_global_zero:
            self._log_epoch_info("test")

    @override
    def on_predict_epoch_start(
        self, trainer: "Trainer", pl_module: "LightningModule"
    ) -> None:
        self._start_time["predict"] = time.monotonic()
        self._total_batches["predict"] = 0

    @override
    def on_predict_batch_end(
        self,
        trainer: "Trainer",
        pl_module: "LightningModule",
        outputs: STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        self._total_batches["predict"] += 1

    @override
    def on_predict_epoch_end(
        self, trainer: "Trainer", pl_module: "LightningModule"
    ) -> None:
        self._elapsed_time["predict"] = time.monotonic() - self._start_time["predict"]
        if trainer.is_global_zero:
            self._log_epoch_info("predict")

    def _log_epoch_info(self, stage: str) -> None:
        if (elapsed_time := self._elapsed_time.get(stage)) is None:
            return
        total_batches = self._total_batches[stage]
        log.critical(
            f"Epoch {stage.capitalize()} Summary: Elapsed Time: {elapsed_time:.2f} seconds | "
            f"Total Batches: {total_batches}"
        )

    @override
    def state_dict(self) -> dict[str, Any]:
        return {
            "elapsed_time": self._elapsed_time,
            "total_batches": self._total_batches,
        }

    @override
    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        self._elapsed_time = state_dict["elapsed_time"]
        self._total_batches = state_dict["total_batches"]


class EpochTimerConfig(CallbackConfigBase):
    name: Literal["epoch_timer"] = "epoch_timer"

    @override
    def create_callbacks(self, root_config):
        yield EpochTimer()



from collections.abc import Iterable, Iterator
from typing import Generic, TypeVar, overload

import torch.nn as nn
from typing_extensions import override

TModule = TypeVar("TModule", bound=nn.Module)


class TypedModuleList(nn.ModuleList, Generic[TModule]):
    def __init__(self, modules: Iterable[TModule] | None = None) -> None:
        super().__init__(modules)

    @overload
    def __getitem__(self, idx: slice) -> "TypedModuleList[TModule]": ...

    @overload
    def __getitem__(self, idx: int) -> TModule: ...

    @override
    def __getitem__(self, idx: int | slice) -> TModule | "TypedModuleList[TModule]":
        return super().__getitem__(idx)  # type: ignore

    @override
    def __setitem__(self, idx: int, module: TModule) -> None:  # type: ignore
        return super().__setitem__(idx, module)

    @override
    def __iter__(self) -> Iterator[TModule]:
        return super().__iter__()  # type: ignore

    @override
    def __iadd__(self, modules: Iterable[TModule]) -> "TypedModuleList[TModule]":  # type: ignore
        return super().__iadd__(modules)  # type: ignore

    @override
    def __add__(self, modules: Iterable[TModule]) -> "TypedModuleList[TModule]":  # type: ignore
        return super().__add__(modules)  # type: ignore

    @override
    def insert(self, idx: int, module: TModule) -> None:  # type: ignore
        return super().insert(idx, module)

    @override
    def append(self, module: TModule) -> "TypedModuleList[TModule]":  # type: ignore
        return super().append(module)  # type: ignore

    @override
    def extend(self, modules: Iterable[TModule]) -> "TypedModuleList[TModule]":  # type: ignore
        return super().extend(modules)  # type: ignore



import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generic, Literal

import numpy as np
import torch
from lightning.pytorch import Trainer
from lightning.pytorch.callbacks import Checkpoint
from typing_extensions import TypeVar, override

from ..._checkpoint.metadata import CheckpointMetadata
from ..._checkpoint.saver import _link_checkpoint, _remove_checkpoint
from ..base import CallbackConfigBase

if TYPE_CHECKING:
    from ...model.config import BaseConfig

log = logging.getLogger(__name__)


class BaseCheckpointCallbackConfig(CallbackConfigBase, ABC):
    dirpath: str | Path | None = None
    """Directory path to save the checkpoint file."""

    filename: str | None = None
    """Checkpoint filename. This must not include the extension.
    If None, the default filename will be used."""

    save_weights_only: bool = False
    """Whether to save only the model's weights or the entire model object."""

    save_symlink: bool = True
    """Whether to create a symlink to the saved checkpoint."""

    topk: int | Literal["all"] = 1
    """The number of checkpoints to keep."""

    @abstractmethod
    def create_checkpoint(
        self,
        root_config: "BaseConfig",
        dirpath: Path,
    ) -> "CheckpointBase | None": ...

    @override
    def create_callbacks(self, root_config):
        dirpath = Path(
            self.dirpath
            or root_config.directory.resolve_subdirectory(root_config.id, "checkpoint")
        )

        if (callback := self.create_checkpoint(root_config, dirpath)) is not None:
            yield callback


TConfig = TypeVar("TConfig", bound=BaseCheckpointCallbackConfig, infer_variance=True)


class CheckpointBase(Checkpoint, ABC, Generic[TConfig]):
    def __init__(self, config: TConfig, dirpath: Path):
        super().__init__()

        self.config = config
        self.dirpath = dirpath / self.name()
        self.dirpath.mkdir(parents=True, exist_ok=True)
        self.symlink_dirpath = dirpath

    @abstractmethod
    def default_filename(self) -> str: ...

    @abstractmethod
    def name(self) -> str: ...

    def extension(self) -> str:
        return ".ckpt"

    @abstractmethod
    def topk_sort_key(self, metadata: CheckpointMetadata) -> Any: ...

    @abstractmethod
    def topk_sort_reverse(self) -> bool: ...

    def symlink_path(self):
        if not self.config.save_symlink:
            return None

        return self.symlink_dirpath / f"{self.name()}{self.extension()}"

    def resolve_checkpoint_path(self, current_metrics: dict[str, Any]) -> Path:
        if (filename := self.config.filename) is None:
            filename = self.default_filename()
        filename = filename.format(**current_metrics)
        return self.dirpath / f"{filename}{self.extension()}"

    def remove_old_checkpoints(self, trainer: Trainer):
        if (topk := self.config.topk) == "all":
            return

        # Get all the checkpoint metadata
        metas = [
            CheckpointMetadata.from_file(p)
            for p in self.dirpath.glob(f"*{CheckpointMetadata.PATH_SUFFIX}")
            if p.is_file() and not p.is_symlink()
        ]

        # Sort by the topk sort key
        metas = sorted(metas, key=self.topk_sort_key, reverse=self.topk_sort_reverse())

        # Now, the metas are sorted from the best to the worst,
        # so we can remove the worst checkpoints
        for meta in metas[topk:]:
            if not (old_ckpt_path := self.dirpath / meta.checkpoint_filename).exists():
                log.warning(
                    f"Checkpoint file not found: {old_ckpt_path}\n"
                    "Skipping removal of the checkpoint metadata."
                )
                continue

            _remove_checkpoint(trainer, old_ckpt_path, metadata=True)
            log.debug(f"Removed old checkpoint: {old_ckpt_path}")

    def current_metrics(self, trainer: Trainer) -> dict[str, Any]:
        current_metrics: dict[str, Any] = {
            "epoch": trainer.current_epoch,
            "step": trainer.global_step,
        }

        for name, value in trainer.callback_metrics.items():
            match value:
                case torch.Tensor() if value.numel() == 1:
                    value = value.detach().cpu().item()
                case np.ndarray() if value.size == 1:
                    value = value.item()
                case _:
                    pass

            current_metrics[name] = value

        return current_metrics

    def save_checkpoints(self, trainer: Trainer):
        if self._should_skip_saving_checkpoint(trainer):
            return

        from ...trainer import Trainer as NTTrainer

        if not isinstance(trainer, NTTrainer):
            raise TypeError(
                f"Trainer must be an instance of {NTTrainer.__name__}, "
                f"but got {type(trainer).__name__}"
            )

        # Save the new checkpoint
        filepath = self.resolve_checkpoint_path(self.current_metrics(trainer))
        trainer.save_checkpoint(filepath, self.config.save_weights_only)

        if trainer.is_global_zero:
            # Remove old checkpoints
            self.remove_old_checkpoints(trainer)

            # Create the latest symlink
            if (symlink_filename := self.symlink_path()) is not None:
                symlink_path = self.dirpath / symlink_filename
                _link_checkpoint(filepath, symlink_path, metadata=True)
                log.debug(f"Created latest symlink: {symlink_path}")

        # Barrier to ensure all processes have saved the checkpoint,
        # deleted the old checkpoints, and created the symlink before continuing
        trainer.strategy.barrier()

    def _should_skip_saving_checkpoint(self, trainer: Trainer) -> bool:
        from lightning.pytorch.trainer.states import TrainerFn

        return (
            bool(
                getattr(trainer, "fast_dev_run", False)
            )  # disable checkpointing with fast_dev_run
            or trainer.state.fn
            != TrainerFn.FITTING  # don't save anything during non-fit
            or trainer.sanity_checking  # don't save anything during sanity check
        )



import logging
from pathlib import Path

import nshconfig as C

from .callbacks.directory_setup import DirectorySetupConfig
from .loggers import LoggerConfig

log = logging.getLogger(__name__)


class DirectoryConfig(C.Config):
    project_root: Path | None = None
    """
    Root directory for this project.

    This isn't specific to the run; it is the parent directory of all runs.
    """

    log: Path | None = None
    """Base directory for all experiment tracking (e.g., WandB, Tensorboard, etc.) files. If None, will use nshtrainer/{id}/log/."""

    stdio: Path | None = None
    """stdout/stderr log directory to use for the trainer. If None, will use nshtrainer/{id}/stdio/."""

    checkpoint: Path | None = None
    """Checkpoint directory to use for the trainer. If None, will use nshtrainer/{id}/checkpoint/."""

    activation: Path | None = None
    """Activation directory to use for the trainer. If None, will use nshtrainer/{id}/activation/."""

    profile: Path | None = None
    """Directory to save profiling information to. If None, will use nshtrainer/{id}/profile/."""

    setup_callback: DirectorySetupConfig = DirectorySetupConfig()
    """Configuration for the directory setup PyTorch Lightning callback."""

    def resolve_run_root_directory(self, run_id: str) -> Path:
        if (project_root_dir := self.project_root) is None:
            project_root_dir = Path.cwd()

        # The default base dir is $CWD/nshtrainer/{id}/
        base_dir = project_root_dir / "nshtrainer"
        base_dir.mkdir(exist_ok=True)

        # Add a .gitignore file to the nshtrainer directory
        #   which will ignore all files except for the .gitignore file itself
        gitignore_path = base_dir / ".gitignore"
        if not gitignore_path.exists():
            gitignore_path.touch()
            gitignore_path.write_text("*\n")

        base_dir = base_dir / run_id
        base_dir.mkdir(exist_ok=True)

        return base_dir

    def resolve_subdirectory(
        self,
        run_id: str,
        # subdirectory: Literal["log", "stdio", "checkpoint", "activation", "profile"],
        subdirectory: str,
    ) -> Path:
        # The subdir will be $CWD/nshtrainer/{id}/{log, stdio, checkpoint, activation}/
        if (subdir := getattr(self, subdirectory, None)) is not None:
            assert isinstance(
                subdir, Path
            ), f"Expected a Path for {subdirectory}, got {type(subdir)}"
            return subdir

        dir = self.resolve_run_root_directory(run_id)
        dir = dir / subdirectory
        dir.mkdir(exist_ok=True)
        return dir

    def _resolve_log_directory_for_logger(self, run_id: str, logger: LoggerConfig):
        if (log_dir := logger.log_dir) is not None:
            return log_dir

        # Save to nshtrainer/{id}/log/{logger name}
        log_dir = self.resolve_subdirectory(run_id, "log")
        log_dir = log_dir / logger.name
        log_dir.mkdir(exist_ok=True)

        return log_dir



from collections import deque
from collections.abc import Callable, Generator
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import torchmetrics
from lightning.pytorch import LightningModule
from lightning.pytorch.utilities.types import _METRIC
from lightning_utilities.core.rank_zero import rank_zero_warn
from nshutils import ActSave
from typing_extensions import override

from ...util.typing_utils import mixin_base_type
from ..config import BaseConfig


@dataclass(frozen=True, kw_only=True)
class _LogContext:
    prefix: str | None = None
    disabled: bool | None = None
    kwargs: dict[str, Any] = field(default_factory=dict)


class LoggerModuleMixin(mixin_base_type(LightningModule)):
    @property
    def log_dir(self):
        """
        The directory where logs are saved.
        """
        if (trainer := self._trainer) is None:
            raise RuntimeError("trainer is not defined")

        if (logger := trainer.logger) is None:
            raise RuntimeError("trainer.logger is not defined")

        if (log_dir := logger.log_dir) is None:
            raise RuntimeError("trainer.logger.log_dir is not defined")

        return Path(log_dir)

    @property
    def should_update_logs(self):
        """
        Whether logs should be updated. This is true once every `log_every_n_steps` steps.
        """
        if self._trainer is None:
            raise RuntimeError(
                "`should_update_logs` can only be used after the module is attached to a trainer"
            )

        return self._trainer._logger_connector.should_update_logs


class LoggerLightningModuleMixin(LoggerModuleMixin, mixin_base_type(LightningModule)):
    @override
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._logger_prefix_stack = deque[_LogContext]()

    if TYPE_CHECKING:

        @contextmanager
        def log_context(
            self,
            prefix: str | None = None,
            *,
            disabled: bool | None = None,
            prog_bar: bool | None = None,
            logger: bool | None = None,
            on_step: bool | None = None,
            on_epoch: bool | None = None,
            reduce_fx: str | Callable | None = None,
            enable_graph: bool | None = None,
            sync_dist: bool | None = None,
            sync_dist_group: Any | None = None,
            add_dataloader_idx: bool | None = None,
            batch_size: int | None = None,
            rank_zero_only: bool | None = None,
        ) -> Generator[None, None, None]: ...

    else:

        @contextmanager
        def log_context(
            self, prefix: str | None = None, *, disabled: bool | None = None, **kwargs
        ) -> Generator[None, None, None]:
            self._logger_prefix_stack.append(
                _LogContext(
                    prefix=prefix,
                    disabled=disabled,
                    kwargs=kwargs,
                )
            )
            try:
                yield
            finally:
                _ = self._logger_prefix_stack.pop()

    if TYPE_CHECKING:

        @override
        def log(  # type: ignore[override]
            self,
            name: str,
            value: _METRIC,
            *,
            prog_bar: bool = False,
            logger: bool | None = None,
            on_step: bool | None = None,
            on_epoch: bool | None = None,
            reduce_fx: str | Callable = "mean",
            enable_graph: bool = False,
            sync_dist: bool = False,
            sync_dist_group: Any | None = None,
            add_dataloader_idx: bool = True,
            batch_size: int | None = None,
            metric_attribute: str | None = None,
            rank_zero_only: bool = False,
        ) -> None: ...

    else:

        @override
        def log(self, name: str, value: _METRIC, **kwargs) -> None:
            # join all prefixes
            prefix = "".join(c.prefix for c in self._logger_prefix_stack if c.prefix)
            name = f"{prefix}{name}"

            # check for disabled context:
            # if the topmost non-null context is disabled, then we don't log
            for c in reversed(self._logger_prefix_stack):
                if c.disabled is not None:
                    if c.disabled:
                        rank_zero_warn(
                            f"Skipping logging of {name} due to disabled context"
                        )
                        return
                    else:
                        break

            fn_kwargs = {}
            for c in self._logger_prefix_stack:
                fn_kwargs.update(c.kwargs)
            fn_kwargs.update(kwargs)

            self._logger_actsave(name, value)

            return super().log(name, value, **fn_kwargs)

    def _logger_actsave(self, name: str, value: _METRIC) -> None:
        hparams = cast(BaseConfig, self.hparams)
        if not hparams.trainer.logging.actsave_logged_metrics:
            return

        ActSave.save(
            lambda: {
                f"logger.{name}": lambda: value.compute()
                if isinstance(value, torchmetrics.Metric)
                else value
            }
        )



import logging
import os
from pathlib import Path
from typing import Literal, cast

from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.callbacks.callback import Callback
from lightning.pytorch.loggers import WandbLogger
from nshrunner._env import SNAPSHOT_DIR
from typing_extensions import override

from .base import CallbackConfigBase

log = logging.getLogger(__name__)


class WandbUploadCodeConfig(CallbackConfigBase):
    name: Literal["wandb_upload_code"] = "wandb_upload_code"

    enabled: bool = True
    """Enable uploading the code to wandb."""

    def __bool__(self):
        return self.enabled

    @override
    def create_callbacks(self, root_config):
        if not self:
            return

        yield WandbUploadCodeCallback(self)


class WandbUploadCodeCallback(Callback):
    def __init__(self, config: WandbUploadCodeConfig):
        super().__init__()

        self.config = config

    @override
    def setup(self, trainer: Trainer, pl_module: LightningModule, stage: str):
        if not self.config:
            return

        if not trainer.is_global_zero:
            return

        if (
            logger := next(
                (
                    logger
                    for logger in trainer.loggers
                    if isinstance(logger, WandbLogger)
                ),
                None,
            )
        ) is None:
            log.warning("Wandb logger not found. Skipping code upload.")
            return

        from wandb.wandb_run import Run

        run = cast(Run, logger.experiment)

        # If a snapshot has been taken (which can be detected using the SNAPSHOT_DIR env),
        # then upload all contents within the snapshot directory to the repository.
        if not (snapshot_dir := os.environ.get(SNAPSHOT_DIR)):
            log.debug("No snapshot directory found. Skipping upload.")
            return

        snapshot_dir = Path(snapshot_dir)
        if not snapshot_dir.exists() or not snapshot_dir.is_dir():
            log.warning(
                f"Snapshot directory '{snapshot_dir}' does not exist or is not a directory."
            )
            return

        log.info(f"Uploading code from snapshot directory '{snapshot_dir}'")
        run.log_code(str(snapshot_dir.absolute()))
